{% extends "defindex.html" %}
{% block tables %}
  <p><strong>Parts of the documentation:</strong></p>
  <table class="contentstable" align="center"><tr>
    <td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("user/index") }}">Numpy User Guide</a><br/>
         <span class="linkdescr">start here</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("reference/index") }}">Numpy Reference</a><br/>
         <span class="linkdescr">reference documentation</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("f2py/index") }}">F2Py Guide</a><br/>
         <span class="linkdescr">f2py documentation</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("dev/index") }}">Numpy Developer Guide</a><br/>
         <span class="linkdescr">contributing to NumPy</span></p>
    </td></tr>
  </table>

  <p><strong>Indices and tables:</strong></p>
  <table class="contentstable" align="center"><tr>
    <td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("genindex") }}">General Index</a><br/>
         <span class="linkdescr">all functions, classes, terms</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("glossary") }}">Glossary</a><br/>
         <span class="linkdescr">the most important terms explained</span></p>
    </td><td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("search") }}">Search page</a><br/>
         <span class="linkdescr">search this documentation</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("contents") }}">Complete Table of Contents</a><br/>
         <span class="linkdescr">lists all sections and subsections</span></p>
    </td></tr>
  </table>

  <p><strong>Meta information:</strong></p>
  <table class="contentstable" align="center"><tr>
    <td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("bugs") }}">Reporting bugs</a></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("about") }}">About NumPy</a></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("neps/index") }}">Numpy Enhancement Proposals</a><br/>
    </td><td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("release") }}">Release Notes</a></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("license") }}">License of Numpy</a></p>
    </td></tr>
  </table>

  <h2>Acknowledgements</h2>
  <p>
    Large parts of this manual originate from Travis E. Oliphant's book
    <a href="http://www.tramy.us/">"Guide to Numpy"</a> (which generously entered
    Public Domain in August 2008). The reference documentation for many of
    the functions are written by numerous contributors and developers of
    Numpy, both prior to and during the
    <a href="http://docs.scipy.org/numpy/">Numpy Documentation Marathon</a>.
  </p>
  <p>
    The preferred way to update the documentation is by submitting a pull
    request on Github (see the 
    <a href="http://docs.scipy.org/doc/numpy-dev/dev/">Developer Guide</a>.
    The <a href="http://docs.scipy.org/numpy/">Numpy Documentation Wiki</a>
    can also still be used to submit documentation fixes.
    Please help us to further improve the Numpy documentation!
  </p>
{% endblock %}
