//spc extend settings

body {
	background-color: rgb(249,250,245);
}

.container {
	width: 80%;
}

.main {
  background-color: white;
  padding: 18px;
  -moz-box-shadow: 0 0 3px #888;
  -webkit-box-shadow: 0 0 3px #888;
  box-shadow: 0 0 3px #888;
}

@import "spc-header.less";
@import "spc-content.less";
@import "spc-rightsidebar.less";
@import "spc-footer.less";