/* Styling for the source code listings: (mostly from pygments)*/

.highlight pre{
    overflow: auto;
    padding: 5px;
    background-color: #ffffff;
    color: #333333;
    border: 1px solid #ac9;
    border-left: none;
    border-right: none;
}

/* Styling for pre elements: from http://perishablepress.com/press/2009/11/09/perfect-pre-tags/ */
/* no vertical scrollbars for IE 6 */
* html pre {
	padding-bottom:25px;
	overflow-y:hidden;
	overflow:visible;
	overflow-x:auto
}
/* no vertical scrollbars for IE 7 */
*:first-child+html pre {
	padding-bottom:25px;
	overflow-y:hidden;
	overflow:visible;
	overflow-x:auto
}

div#spc-section-body td.linenos pre {
    padding: 5px 0px;
    border: 0;
    background-color: transparent;
    color: #aaa;
}
.highlight .hll { background-color: #ffffcc }
.highlight  { background: #ffffff; }
.highlight .c { color: #008000 } /* Comment */
.highlight .k { color: #000080; font-weight: bold } /* Keyword */
.highlight .n { color: #000000 } /* Name */
.highlight .o { color: #000000 } /* Operator */
.highlight .cm { color: #008000 } /* Comment.Multiline */
.highlight .cp { color: #008000 } /* Comment.Preproc */
.highlight .c1 { color: #008000 } /* Comment.Single */
.highlight .cs { color: #008000 } /* Comment.Special */
.highlight .kc { color: #000080; font-weight: bold } /* Keyword.Constant */
.highlight .kd { color: #000080; font-weight: bold } /* Keyword.Declaration */
.highlight .kn { color: #000080; font-weight: bold } /* Keyword.Namespace */
.highlight .kp { color: #000080; font-weight: bold } /* Keyword.Pseudo */
.highlight .kr { color: #000080; font-weight: bold } /* Keyword.Reserved */
.highlight .kt { color: #000080; font-weight: bold } /* Keyword.Type */
.highlight .m { color: #008080 } /* Literal.Number */
.highlight .s { color: #800080 } /* Literal.String */
.highlight .na { color: #000000 } /* Name.Attribute */
.highlight .nb { color: #407090 } /* Name.Builtin */
.highlight .nc { color: #0000F0; font-weight: bold } /* Name.Class */
.highlight .no { color: #000000 } /* Name.Constant */
.highlight .nd { color: #000000 } /* Name.Decorator */
.highlight .ni { color: #000000 } /* Name.Entity */
.highlight .ne { color: #000000 } /* Name.Exception */
.highlight .nf { color: #008080; font-weight: bold } /* Name.Function */
.highlight .nl { color: #000000 } /* Name.Label */
.highlight .nn { color: #000000 } /* Name.Namespace */
.highlight .nx { color: #000000 } /* Name.Other */
.highlight .py { color: #000000 } /* Name.Property */
.highlight .nt { color: #000000 } /* Name.Tag */
.highlight .nv { color: #000000 } /* Name.Variable */
.highlight .ow { color: #AA22FF; font-weight: bold } /* Operator.Word */
.highlight .mf { color: #008080 } /* Literal.Number.Float */
.highlight .mh { color: #008080 } /* Literal.Number.Hex */
.highlight .mi { color: #008080 } /* Literal.Number.Integer */
.highlight .mo { color: #008080 } /* Literal.Number.Oct */
.highlight .sb { color: #800080 } /* Literal.String.Backtick */
.highlight .sc { color: #800080 } /* Literal.String.Char */
.highlight .sd { color: #800000 } /* Literal.String.Doc */
.highlight .s2 { color: #800080 } /* Literal.String.Double */
.highlight .se { color: #800080 } /* Literal.String.Escape */
.highlight .sh { color: #800080 } /* Literal.String.Heredoc */
.highlight .si { color: #800080 } /* Literal.String.Interpol */
.highlight .sx { color: #800080 } /* Literal.String.Other */
.highlight .sr { color: #800080 } /* Literal.String.Regex */
.highlight .s1 { color: #800080 } /* Literal.String.Single */
.highlight .ss { color: #800080 } /* Literal.String.Symbol */
.highlight .bp { color: #407090 } /* Name.Builtin.Pseudo */
.highlight .vc { color: #000000 } /* Name.Variable.Class */
.highlight .vg { color: #000000 } /* Name.Variable.Global */
.highlight .vi { color: #000000 } /* Name.Variable.Instance */
.highlight .il { color: #008080 } /* Literal.Number.Integer.Long */
